@@ -1,5 +1,5 @@
 // ======================================================================== //
-// Copyright 2009-2018 Intel Corporation                                    //
+// Copyright 2009-2020 Intel Corporation                                    //
 //                                                                          //
 // Licensed under the Apache License, Version 2.0 (the "License");          //
 // you may not use this file except in compliance with the License.         //
@@ -18,6 +18,11 @@
 #include "intrinsics.h"
 #include "string.h"
 #include "ref.h"
+#if defined(__FREEBSD__)
+#include <sys/cpuset.h>
+#include <pthread_np.h>
+typedef cpuset_t cpu_set_t;
+#endif
 
 ////////////////////////////////////////////////////////////////////////////////
 /// All Platforms
@@ -542,13 +547,24 @@ namespace embree
 #include <unistd.h>
 #include <sys/ioctl.h>
 #include <sys/time.h>
+#include <pthread.h>
 
 namespace embree
 {
   unsigned int getNumberOfLogicalThreads() 
   {
     static int nThreads = -1;
-    if (nThreads == -1) nThreads = sysconf(_SC_NPROCESSORS_CONF);
+    if (nThreads != -1) return nThreads;
+
+#if defined(__MACOSX__)
+    nThreads = sysconf(_SC_NPROCESSORS_ONLN); // does not work in Linux LXC container
+    assert(nThreads);
+#else
+    cpu_set_t set;
+    if (pthread_getaffinity_np(pthread_self(), sizeof(set), &set) == 0)
+      nThreads = CPU_COUNT(&set);
+#endif
+    
     assert(nThreads);
     return nThreads;
   }
