@@ -1,5 +1,5 @@
 // ======================================================================== //
-// Copyright 2009-2018 Intel Corporation                                    //
+// Copyright 2009-2020 Intel Corporation                                    //
 //                                                                          //
 // Licensed under the Apache License, Version 2.0 (the "License");          //
 // you may not use this file except in compliance with the License.         //
@@ -94,7 +94,7 @@ namespace embree
     object_accel_mb_min_leaf_size = 1;
     object_accel_mb_max_leaf_size = 1;
 
-    max_spatial_split_replications = 2.0f;
+    max_spatial_split_replications = 1.2f;
     useSpatialPreSplits = false;
 
     tessellation_cache_size = 128*1024*1024;
@@ -121,6 +121,8 @@ namespace embree
     benchmark = 0;
 
     numThreads = 0;
+    numUserThreads = 0;
+
 #if TASKING_INTERNAL
     set_affinity = true;
 #else
@@ -248,6 +250,9 @@ namespace embree
       if (tok == Token::Id("threads") && cin->trySymbol("=")) 
         numThreads = cin->get().Int();
 
+      else if (tok == Token::Id("user_threads")&& cin->trySymbol("=")) 
+        numUserThreads = cin->get().Int();
+
       else if (tok == Token::Id("set_affinity")&& cin->trySymbol("=")) 
         set_affinity = cin->get().Int();
 
@@ -449,10 +454,11 @@ namespace embree
   void State::print()
   {
     std::cout << "general:" << std::endl;
-    std::cout << "  build threads = " << numThreads   << std::endl;
-    std::cout << "  start_threads = " << start_threads << std::endl;
-    std::cout << "  affinity      = " << set_affinity << std::endl;
-    std::cout << "  frequency_level = ";
+    std::cout << "  build threads      = " << numThreads   << std::endl;
+    std::cout << "  build user threads = " << numUserThreads   << std::endl;
+    std::cout << "  start_threads      = " << start_threads << std::endl;
+    std::cout << "  affinity           = " << set_affinity << std::endl;
+    std::cout << "  frequency_level    = ";
     switch (frequency_level) {
     case FREQUENCY_SIMD128: std::cout << "simd128" << std::endl; break;
     case FREQUENCY_SIMD256: std::cout << "simd256" << std::endl; break;
@@ -460,72 +466,72 @@ namespace embree
     default: std::cout << "error" << std::endl; break;
     }
     
-    std::cout << "  hugepages     = ";
+    std::cout << "  hugepages          = ";
     if (!hugepages) std::cout << "disabled" << std::endl;
     else if (hugepages_success) std::cout << "enabled" << std::endl;
     else std::cout << "failed" << std::endl;
 
-    std::cout << "  verbosity     = " << verbose << std::endl;
-    std::cout << "  cache_size    = " << float(tessellation_cache_size)*1E-6 << " MB" << std::endl;
+    std::cout << "  verbosity          = " << verbose << std::endl;
+    std::cout << "  cache_size         = " << float(tessellation_cache_size)*1E-6 << " MB" << std::endl;
     std::cout << "  max_spatial_split_replications = " << max_spatial_split_replications << std::endl;
     
     std::cout << "triangles:" << std::endl;
-    std::cout << "  accel         = " << tri_accel << std::endl;
-    std::cout << "  builder       = " << tri_builder << std::endl;
-    std::cout << "  traverser     = " << tri_traverser << std::endl;
+    std::cout << "  accel              = " << tri_accel << std::endl;
+    std::cout << "  builder            = " << tri_builder << std::endl;
+    std::cout << "  traverser          = " << tri_traverser << std::endl;
         
     std::cout << "motion blur triangles:" << std::endl;
-    std::cout << "  accel         = " << tri_accel_mb << std::endl;
-    std::cout << "  builder       = " << tri_builder_mb << std::endl;
-    std::cout << "  traverser     = " << tri_traverser_mb << std::endl;
+    std::cout << "  accel              = " << tri_accel_mb << std::endl;
+    std::cout << "  builder            = " << tri_builder_mb << std::endl;
+    std::cout << "  traverser          = " << tri_traverser_mb << std::endl;
 
     std::cout << "quads:" << std::endl;
-    std::cout << "  accel         = " << quad_accel << std::endl;
-    std::cout << "  builder       = " << quad_builder << std::endl;
-    std::cout << "  traverser     = " << quad_traverser << std::endl;
+    std::cout << "  accel              = " << quad_accel << std::endl;
+    std::cout << "  builder            = " << quad_builder << std::endl;
+    std::cout << "  traverser          = " << quad_traverser << std::endl;
 
     std::cout << "motion blur quads:" << std::endl;
-    std::cout << "  accel         = " << quad_accel_mb << std::endl;
-    std::cout << "  builder       = " << quad_builder_mb << std::endl;
-    std::cout << "  traverser     = " << quad_traverser_mb << std::endl;
+    std::cout << "  accel              = " << quad_accel_mb << std::endl;
+    std::cout << "  builder            = " << quad_builder_mb << std::endl;
+    std::cout << "  traverser          = " << quad_traverser_mb << std::endl;
 
     std::cout << "line segments:" << std::endl;
-    std::cout << "  accel         = " << line_accel << std::endl;
-    std::cout << "  builder       = " << line_builder << std::endl;
-    std::cout << "  traverser     = " << line_traverser << std::endl;
+    std::cout << "  accel              = " << line_accel << std::endl;
+    std::cout << "  builder            = " << line_builder << std::endl;
+    std::cout << "  traverser          = " << line_traverser << std::endl;
 
     std::cout << "motion blur line segments:" << std::endl;
-    std::cout << "  accel         = " << line_accel_mb << std::endl;
-    std::cout << "  builder       = " << line_builder_mb << std::endl;
-    std::cout << "  traverser     = " << line_traverser_mb << std::endl;
+    std::cout << "  accel              = " << line_accel_mb << std::endl;
+    std::cout << "  builder            = " << line_builder_mb << std::endl;
+    std::cout << "  traverser          = " << line_traverser_mb << std::endl;
     
     std::cout << "hair:" << std::endl;
-    std::cout << "  accel         = " << hair_accel << std::endl;
-    std::cout << "  builder       = " << hair_builder << std::endl;
-    std::cout << "  traverser     = " << hair_traverser << std::endl;
+    std::cout << "  accel              = " << hair_accel << std::endl;
+    std::cout << "  builder            = " << hair_builder << std::endl;
+    std::cout << "  traverser          = " << hair_traverser << std::endl;
 
     std::cout << "motion blur hair:" << std::endl;
-    std::cout << "  accel         = " << hair_accel_mb << std::endl;
-    std::cout << "  builder       = " << hair_builder_mb << std::endl;
-    std::cout << "  traverser     = " << hair_traverser_mb << std::endl;
+    std::cout << "  accel              = " << hair_accel_mb << std::endl;
+    std::cout << "  builder            = " << hair_builder_mb << std::endl;
+    std::cout << "  traverser          = " << hair_traverser_mb << std::endl;
     
     std::cout << "subdivision surfaces:" << std::endl;
-    std::cout << "  accel         = " << subdiv_accel << std::endl;
+    std::cout << "  accel              = " << subdiv_accel << std::endl;
 
     std::cout << "grids:" << std::endl;
-    std::cout << "  accel         = " << grid_accel << std::endl;
-    std::cout << "  builder       = " << grid_builder << std::endl;
+    std::cout << "  accel              = " << grid_accel << std::endl;
+    std::cout << "  builder            = " << grid_builder << std::endl;
 
     std::cout << "motion blur grids:" << std::endl;
-    std::cout << "  accel         = " << grid_accel_mb << std::endl;
-    std::cout << "  builder       = " << grid_builder_mb << std::endl;
+    std::cout << "  accel              = " << grid_accel_mb << std::endl;
+    std::cout << "  builder            = " << grid_builder_mb << std::endl;
 
     std::cout << "object_accel:" << std::endl;
-    std::cout << "  min_leaf_size = " << object_accel_min_leaf_size << std::endl;
-    std::cout << "  max_leaf_size = " << object_accel_max_leaf_size << std::endl;
+    std::cout << "  min_leaf_size      = " << object_accel_min_leaf_size << std::endl;
+    std::cout << "  max_leaf_size      = " << object_accel_max_leaf_size << std::endl;
 
     std::cout << "object_accel_mb:" << std::endl;
-    std::cout << "  min_leaf_size = " << object_accel_mb_min_leaf_size << std::endl;
-    std::cout << "  max_leaf_size = " << object_accel_mb_max_leaf_size << std::endl;
+    std::cout << "  min_leaf_size      = " << object_accel_mb_min_leaf_size << std::endl;
+    std::cout << "  max_leaf_size      = " << object_accel_mb_max_leaf_size << std::endl;
   }
 }
