@@ -1,5 +1,5 @@
 // ======================================================================== //
-// Copyright 2009-2018 Intel Corporation                                    //
+// Copyright 2009-2020 Intel Corporation                                    //
 //                                                                          //
 // Licensed under the Apache License, Version 2.0 (the "License");          //
 // you may not use this file except in compliance with the License.         //
@@ -37,18 +37,14 @@ namespace embree
       __forceinline QuadHitPlueckerM(const vbool<M>& valid,
                                      const vfloat<M>& U,
                                      const vfloat<M>& V,
-                                     const vfloat<M>& W,
-                                     const vfloat<M>& T,
-                                     const vfloat<M>& absDen,
+                                     const vfloat<M>& UVW,
+                                     const vfloat<M>& t,
                                      const Vec3vf<M>& Ng,
                                      const vbool<M>& flags)
-        : U(U), V(V), W(W), T(T), absDen(absDen), tri_Ng(Ng), valid(valid), flags(flags) {}
+        : U(U), V(V), UVW(UVW), tri_Ng(Ng), valid(valid), vt(t), flags(flags) {}
 
       __forceinline void finalize()
       {
-        const vfloat<M> rcpAbsDen = rcp(absDen);
-        vt = T * rcpAbsDen;
-        const vfloat<M> UVW = U+V+W;
         const vbool<M> invalid = abs(UVW) < min_rcp_input;
         const vfloat<M> rcpUVW = select(invalid,vfloat<M>(0.0f),rcp(UVW));
         const vfloat<M> u = U * rcpUVW;
@@ -80,9 +76,7 @@ namespace embree
     private:
       vfloat<M> U;
       vfloat<M> V;
-      vfloat<M> W;
-      vfloat<M> T;
-      vfloat<M> absDen;
+      vfloat<M> UVW;
       Vec3vf<M> tri_Ng;
 
     public:
@@ -101,18 +95,14 @@ namespace embree
     {
       __forceinline QuadHitPlueckerK(const vfloat<K>& U,
                                      const vfloat<K>& V,
-                                     const vfloat<K>& W,
-                                     const vfloat<K>& T,
-                                     const vfloat<K>& absDen,
+                                     const vfloat<K>& UVW,
+                                     const vfloat<K>& t,
                                      const Vec3vf<K>& Ng,
                                      const vbool<K>& flags)
-        : U(U), V(V), W(W), T(T), absDen(absDen), flags(flags), tri_Ng(Ng) {}
+        : U(U), V(V), UVW(UVW), t(t), flags(flags), tri_Ng(Ng) {}
 
       __forceinline std::tuple<vfloat<K>,vfloat<K>,vfloat<K>,Vec3vf<K>> operator() () const
       {
-        const vfloat<K> rcpAbsDen = rcp(absDen);
-        const vfloat<K> t = T * rcpAbsDen;
-        const vfloat<K> UVW = U+V+W;
         const vbool<K> invalid = abs(UVW) < min_rcp_input;
         const vfloat<K> rcpUVW = select(invalid,vfloat<K>(0.0f),rcp(UVW));
         const vfloat<K> u0 = U * rcpUVW;
@@ -128,9 +118,8 @@ namespace embree
     private:
       const vfloat<K> U;
       const vfloat<K> V;
-      const vfloat<K> W;
-      const vfloat<K> T;
-      const vfloat<K> absDen;
+      const vfloat<K> UVW;
+      const vfloat<K> t;
       const vbool<K> flags;
       const Vec3vf<K> tri_Ng;
     };
@@ -161,35 +150,28 @@ namespace embree
         const vfloat<M> U = dot(cross(e0,v2+v0),D);
         const vfloat<M> V = dot(cross(e1,v0+v1),D);
         const vfloat<M> W = dot(cross(e2,v1+v2),D);
-        const vfloat<M> eps = float(ulp)*abs(U+V+W);
+        const vfloat<M> UVW = U+V+W;
+        const vfloat<M> eps = float(ulp)*abs(UVW);
 #if defined(EMBREE_BACKFACE_CULLING)
-        const vfloat<M> maxUVW = max(U,V,W);
-        vbool<M> valid = maxUVW <= eps;
+        vbool<M> valid = max(U,V,W) <= eps;
 #else
-        const vfloat<M> minUVW = min(U,V,W);
-        const vfloat<M> maxUVW = max(U,V,W);
-        vbool<M> valid =  (minUVW >= -eps) | (maxUVW <= eps);
+        vbool<M> valid =  (min(U,V,W) >= -eps) | (max(U,V,W) <= eps);
 #endif
         if (unlikely(none(valid))) return false;
 
         /* calculate geometry normal and denominator */
         const Vec3vf<M> Ng = stable_triangle_normal(e0,e1,e2);
         const vfloat<M> den = twice(dot(Ng,D));
-        const vfloat<M> absDen = abs(den);
-        const vfloat<M> sgnDen = signmsk(den);
 
-        /* perform depth test */
+         /* perform depth test */
         const vfloat<M> T = twice(dot(v0,Ng));
-        valid &= absDen*vfloat<M>(ray.tnear()) < (T^sgnDen);
-        valid &= (T^sgnDen) <= absDen*vfloat<M>(ray.tfar);
-        if (unlikely(none(valid))) return false;
-
-        /* avoid division by 0 */
+        const vfloat<M> t = rcp(den)*T;
+        valid &= vfloat<M>(ray.tnear()) <= t & t <= vfloat<M>(ray.tfar);
         valid &= den != vfloat<M>(zero);
         if (unlikely(none(valid))) return false;
 
         /* update hit information */
-        QuadHitPlueckerM<M> hit(valid,U,V,W,T,den,Ng,flags);
+        QuadHitPlueckerM<M> hit(valid,U,V,UVW,t,Ng,flags);
         return epilog(valid,hit);
       }
     };
@@ -341,34 +323,31 @@ namespace embree
           const vfloat<M> U = dot(cross(e0,v2+v0),D);
           const vfloat<M> V = dot(cross(e1,v0+v1),D);
           const vfloat<M> W = dot(cross(e2,v1+v2),D);
-          const vfloat<M> minUVW MAYBE_UNUSED = min(U,V,W);
-          const vfloat<M> maxUVW = max(U,V,W);
-          const vfloat<M> eps = float(ulp)*abs(U+V+W);
+          const vfloat<M> UVW = U+V+W;
+          const vfloat<M> eps = float(ulp)*abs(UVW);
 #if defined(EMBREE_BACKFACE_CULLING)
-          vbool<M> valid = maxUVW <= eps;
+          vbool<M> valid = max(U,V,W) <= eps;
 #else
-          vbool<M> valid = (minUVW >= -eps) | (maxUVW <= eps);
+          vbool<M> valid = (min(U,V,W) >= -eps) | (max(U,V,W) <= eps);
 #endif
           if (unlikely(none(valid))) return false;
           
           /* calculate geometry normal and denominator */
           const Vec3vf<M> Ng = stable_triangle_normal(e0,e1,e2);
           const vfloat<M> den = twice(dot(Ng,D));
-          const vfloat<M> absDen = abs(den);
-          const vfloat<M> sgnDen = signmsk(den);
 
           /* perform depth test */
           const vfloat<M> T = twice(dot(v0,Ng));
-          valid &= absDen*vfloat<M>(ray.tnear()[k]) < (T^sgnDen);
-          valid &= (T^sgnDen) <= absDen*vfloat<M>(ray.tfar[k]);
+          const vfloat<M> t = rcp(den)*T;
+          valid &= vfloat<M>(ray.tnear()[k]) <= t & t <= vfloat<M>(ray.tfar[k]);
           if (unlikely(none(valid))) return false;
           
           /* avoid division by 0 */
           valid &= den != vfloat<M>(zero);
           if (unlikely(none(valid))) return false;
           
-          /* calculate hit information */
-          QuadHitPlueckerM<M> hit(valid,U,V,W,T,den,Ng,flags);
+          /* update hit information */
+          QuadHitPlueckerM<M> hit(valid,U,V,UVW,t,Ng,flags);
           return epilog(valid,hit);
       }
     };
@@ -405,35 +384,28 @@ namespace embree
           const vfloat<K> U = dot(Vec3vf<K>(cross(e0,v2+v0)),D);
           const vfloat<K> V = dot(Vec3vf<K>(cross(e1,v0+v1)),D);
           const vfloat<K> W = dot(Vec3vf<K>(cross(e2,v1+v2)),D);
-          const vfloat<K> eps = float(ulp)*abs(U+V+W);
+          const vfloat<K> UVW = U+V+W;
+          const vfloat<K> eps = float(ulp)*abs(UVW);
 #if defined(EMBREE_BACKFACE_CULLING)
-          const vfloat<K> maxUVW = max(U,V,W);
-          valid &= maxUVW <= eps;
+          valid &= max(U,V,W) <= eps;
 #else
-          const vfloat<K> minUVW = min(U,V,W);
-          const vfloat<K> maxUVW = max(U,V,W);
-          valid &= (minUVW >= -eps) | (maxUVW <= eps);
+          valid &= (min(U,V,W) >= -eps) | (max(U,V,W) <= eps);
 #endif
           if (unlikely(none(valid))) return false;
           
            /* calculate geometry normal and denominator */
           const Vec3vf<K> Ng = stable_triangle_normal(e0,e1,e2);
           const vfloat<K> den = twice(dot(Vec3vf<K>(Ng),D));
-          const vfloat<K> absDen = abs(den);
-          const vfloat<K> sgnDen = signmsk(den);
 
           /* perform depth test */
           const vfloat<K> T = twice(dot(v0,Vec3vf<K>(Ng)));
-          valid &= absDen*ray.tnear() < (T^sgnDen);
-          valid &= (T^sgnDen) <= absDen*ray.tfar;
-          if (unlikely(none(valid))) return false;
-          
-          /* avoid division by 0 */
+          const vfloat<K> t = rcp(den)*T;
+          valid &= ray.tnear() <= t & t <= ray.tfar;
           valid &= den != vfloat<K>(zero);
           if (unlikely(none(valid))) return false;
           
           /* calculate hit information */
-          QuadHitPlueckerK<K> hit(U,V,W,T,den,Ng,flags);
+          QuadHitPlueckerK<K> hit(U,V,UVW,t,Ng,flags);
           return epilog(valid,hit);
       }
       
