@@ -1,5 +1,5 @@
 // ======================================================================== //
-// Copyright 2009-2018 Intel Corporation                                    //
+// Copyright 2009-2020 Intel Corporation                                    //
 //                                                                          //
 // Licensed under the Apache License, Version 2.0 (the "License");          //
 // you may not use this file except in compliance with the License.         //
@@ -82,7 +82,7 @@ namespace embree
   {
     Ref<SceneGraph::GridMeshNode> mesh = new SceneGraph::GridMeshNode(material,BBox1f(0,1),1);
     mesh->positions[0].resize((width+1)*(height+1));
-    mesh->grids.push_back(SceneGraph::GridMeshNode::Grid(0,width+1,width+1,height+1));
+    mesh->grids.push_back(SceneGraph::GridMeshNode::Grid(0,(unsigned)width+1,(unsigned)width+1,(unsigned)height+1));
 
     for (size_t y=0; y<=height; y++) {
       for (size_t x=0; x<=width; x++) {
@@ -256,7 +256,7 @@ namespace embree
 
     for (size_t i=0; i<6; i++)
     {
-      mesh->grids.push_back(SceneGraph::GridMeshNode::Grid(i*grid_size,N+1,N+1,N+1));
+      mesh->grids.push_back(SceneGraph::GridMeshNode::Grid(unsigned(i*grid_size),unsigned(N+1),unsigned(N+1),unsigned(N+1)));
       Vec3fa p0, dx, dy;
       switch (i) {
       case 0: p0 = Vec3fa(-0.5f,-0.5f,-0.5f); dx = Vec3fa(+1,0, 0); dy = Vec3fa(0,1,0); break;
