@@ -1,5 +1,5 @@
 // ======================================================================== //
-// Copyright 2009-2018 Intel Corporation                                    //
+// Copyright 2009-2020 Intel Corporation                                    //
 //                                                                          //
 // Licensed under the Apache License, Version 2.0 (the "License");          //
 // you may not use this file except in compliance with the License.         //
@@ -140,6 +140,7 @@ namespace embree
     rh.ray.tfar = neg_inf;
     rh.ray.time = 0;
     rh.ray.mask = -1;
+    rh.ray.id = -1;
     rh.hit.Ng_x = 0.0f;
     rh.hit.Ng_y = 0.0f;
     rh.hit.Ng_z = 0.0f;
@@ -147,7 +148,8 @@ namespace embree
     rh.hit.v = 0.0f;
     rh.hit.geomID = -1;
     rh.hit.primID = -1;
-    rh.hit.instID[0] = -1;
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rh.hit.instID[l] = RTC_INVALID_GEOMETRY_ID;
   }
 
   __forceinline RTCRayHit makeRay(const Vec3fa& org, const Vec3fa& dir) 
@@ -157,7 +159,9 @@ namespace embree
     rh.ray.dir_x = dir.x; rh.ray.dir_y = dir.y; rh.ray.dir_z = dir.z;
     rh.ray.tnear = 0.0f; rh.ray.tfar = inf;
     rh.ray.time = 0; rh.ray.mask = -1;
-    rh.hit.geomID = rh.hit.primID = rh.hit.instID[0] = -1;
+    rh.hit.geomID = rh.hit.primID = -1;
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rh.hit.instID[l] = RTC_INVALID_GEOMETRY_ID;
     return rh;
   }
 
@@ -168,7 +172,9 @@ namespace embree
     rh.ray.dir_x = dir.x; rh.ray.dir_y = dir.y; rh.ray.dir_z = dir.z;
     rh.ray.tnear = tnear; rh.ray.tfar = tfar;
     rh.ray.time = 0; rh.ray.mask = -1;
-    rh.hit.geomID = rh.hit.primID = rh.hit.instID[0] = -1;
+    rh.hit.geomID = rh.hit.primID = -1;
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rh.hit.instID[l] = RTC_INVALID_GEOMETRY_ID;
     return rh;
   }
 
@@ -181,7 +187,10 @@ namespace embree
     rh.ray.tfar = inf;
     rh.ray.time = 0; 
     rh.ray.mask = -1;
-    rh.hit.geomID = rh.hit.primID = rh.hit.instID[0] = -1;
+    rh.ray.id = -1;
+    rh.hit.geomID = rh.hit.primID = -1;
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rh.hit.instID[l] = RTC_INVALID_GEOMETRY_ID;
     return rh;
   }
 
@@ -198,6 +207,7 @@ namespace embree
     rh.ray.tfar = inf;
     rh.ray.time = 0; 
     rh.ray.mask = -1;
+    rh.ray.id = -1;
     rh.hit.geomID = rh.hit.primID = rh.hit.instID[0] = -1;
   }
 
@@ -212,8 +222,10 @@ namespace embree
     rh.ray.org_x = org.x; rh.ray.org_y = org.y; rh.ray.org_z = org.z; // FIXME: optimize
     rh.ray.dir_x = dir.x; rh.ray.dir_y = dir.y; rh.ray.dir_z = dir.z;
     rh.ray.tnear = tnear; rh.ray.tfar = tfar;
-    rh.ray.time = 0; rh.ray.mask = -1;
-    rh.hit.geomID = rh.hit.primID = rh.hit.instID[0] = -1;
+    rh.ray.time = 0; rh.ray.mask = -1; rh.ray.id = -1;
+    rh.hit.geomID = rh.hit.primID = -1;
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rh.hit.instID[l] = RTC_INVALID_GEOMETRY_ID;
     return rh;
   }
 
@@ -229,9 +241,10 @@ namespace embree
     if (*(int*)&ray0.ray.tfar   != *(int*)&ray1.ray.tfar  ) return true;
     if (*(int*)&ray0.ray.time   != *(int*)&ray1.ray.time  ) return true;
     if (*(int*)&ray0.ray.mask   != *(int*)&ray1.ray.mask  ) return true;
+    if (*(int*)&ray0.ray.id     != *(int*)&ray1.ray.id    ) return true;
     if (*(int*)&ray0.hit.u      != *(int*)&ray1.hit.u     ) return true;
     if (*(int*)&ray0.hit.v      != *(int*)&ray1.hit.v     ) return true;
-    if (*(int*)&ray0.hit.instID[0] != *(int*)&ray1.hit.instID[0]) return true;
+    if (*(int*)&ray0.hit.instID != *(int*)&ray1.hit.instID) return true;
     if (*(int*)&ray0.hit.geomID != *(int*)&ray1.hit.geomID) return true;
     if (*(int*)&ray0.hit.primID != *(int*)&ray1.hit.primID) return true;
     if (*(int*)&ray0.hit.Ng_x  != *(int*)&ray1.hit.Ng_x ) return true;
@@ -243,20 +256,29 @@ namespace embree
   /* Outputs ray to stream */
   __forceinline std::ostream& operator<<(std::ostream& cout, const RTCRayHit& rh)
   {
-    return cout << "Ray { " << std::endl
-                << "  org = " << rh.ray.org_x << " " << rh.ray.org_y << " " << rh.ray.org_z << std::endl
-                << "  dir = " << rh.ray.dir_x << " " << rh.ray.dir_y << " " << rh.ray.dir_z << std::endl
-                << "  near = " << rh.ray.tnear << std::endl
-                << "  far = " << rh.ray.tfar << std::endl
-                << "  time = " << rh.ray.time << std::endl
-                << "  mask = " << rh.ray.mask << std::endl
-                << "  instID = " << rh.hit.instID[0] << std::endl
-                << "  geomID = " << rh.hit.geomID << std::endl
-                << "  primID = " << rh.hit.primID <<  std::endl
-                << "  u = " << rh.hit.u <<  std::endl
-                << "  v = " << rh.hit.v << std::endl
-                << "  Ng = " << rh.hit.Ng_x << " " << rh.hit.Ng_y << " " << rh.hit.Ng_z << std::endl
-                << "}";
+    cout << "Ray { " << std::endl
+         << "  org = " << rh.ray.org_x << " " << rh.ray.org_y << " " << rh.ray.org_z << std::endl
+         << "  dir = " << rh.ray.dir_x << " " << rh.ray.dir_y << " " << rh.ray.dir_z << std::endl
+         << "  near = " << rh.ray.tnear << std::endl
+         << "  far = " << rh.ray.tfar << std::endl
+         << "  time = " << rh.ray.time << std::endl
+         << "  mask = " << rh.ray.mask << std::endl
+         << "  id = " << rh.ray.id << std::endl
+         << "  instID =";
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+    {
+      cout << " " << rh.hit.instID[l];
+      if (rh.hit.instID[l] == RTC_INVALID_GEOMETRY_ID)
+        break;
+    }
+    cout << std::endl
+         << "  geomID = " << rh.hit.geomID << std::endl
+         << "  primID = " << rh.hit.primID <<  std::endl
+         << "  u = " << rh.hit.u <<  std::endl
+         << "  v = " << rh.hit.v << std::endl
+         << "  Ng = " << rh.hit.Ng_x << " " << rh.hit.Ng_y << " " << rh.hit.Ng_z << std::endl
+         << "}";
+    return cout;
   }
 
   __forceinline void setRay(RTCRayHit4& ray_o, size_t i, const RTCRayHit& ray_i)
@@ -271,7 +293,7 @@ namespace embree
     ray_o.ray.tfar[i] = ray_i.ray.tfar;
     ray_o.ray.time[i] = ray_i.ray.time;
     ray_o.ray.mask[i] = ray_i.ray.mask;
-    ray_o.hit.instID[0][i] = ray_i.hit.instID[0];
+    ray_o.ray.id[i] = ray_i.ray.id;
     ray_o.hit.geomID[i] = ray_i.hit.geomID;
     ray_o.hit.primID[i] = ray_i.hit.primID;
     ray_o.hit.u[i] = ray_i.hit.u;
@@ -279,6 +301,9 @@ namespace embree
     ray_o.hit.Ng_x[i] = ray_i.hit.Ng_x;
     ray_o.hit.Ng_y[i] = ray_i.hit.Ng_y;
     ray_o.hit.Ng_z[i] = ray_i.hit.Ng_z;
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l][i] = ray_i.hit.instID[l];
   }
 
   __forceinline void setRay(RTCRayHit8& ray_o, size_t i, const RTCRayHit& ray_i)
@@ -293,7 +318,7 @@ namespace embree
     ray_o.ray.tfar[i] = ray_i.ray.tfar;
     ray_o.ray.time[i] = ray_i.ray.time;
     ray_o.ray.mask[i] = ray_i.ray.mask;
-    ray_o.hit.instID[0][i] = ray_i.hit.instID[0];
+    ray_o.ray.id[i] = ray_i.ray.id;
     ray_o.hit.geomID[i] = ray_i.hit.geomID;
     ray_o.hit.primID[i] = ray_i.hit.primID;
     ray_o.hit.u[i] = ray_i.hit.u;
@@ -301,6 +326,9 @@ namespace embree
     ray_o.hit.Ng_x[i] = ray_i.hit.Ng_x;
     ray_o.hit.Ng_y[i] = ray_i.hit.Ng_y;
     ray_o.hit.Ng_z[i] = ray_i.hit.Ng_z;
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l][i] = ray_i.hit.instID[l];
   }
 
   __forceinline void setRay(RTCRayHit16& ray_o, size_t i, const RTCRayHit& ray_i)
@@ -315,7 +343,7 @@ namespace embree
     ray_o.ray.tfar[i] = ray_i.ray.tfar;
     ray_o.ray.time[i] = ray_i.ray.time;
     ray_o.ray.mask[i] = ray_i.ray.mask;
-    ray_o.hit.instID[0][i] = ray_i.hit.instID[0];
+    ray_o.ray.id[i] = ray_i.ray.id;
     ray_o.hit.geomID[i] = ray_i.hit.geomID;
     ray_o.hit.primID[i] = ray_i.hit.primID;
     ray_o.hit.u[i] = ray_i.hit.u;
@@ -323,6 +351,9 @@ namespace embree
     ray_o.hit.Ng_x[i] = ray_i.hit.Ng_x;
     ray_o.hit.Ng_y[i] = ray_i.hit.Ng_y;
     ray_o.hit.Ng_z[i] = ray_i.hit.Ng_z;
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l][i] = ray_i.hit.instID[l];
   }
 
   __forceinline void setRay(RTCRayHitN* rayhit_o, unsigned int N, unsigned int i, const RTCRayHit& ray_i)
@@ -338,8 +369,8 @@ namespace embree
     RTCRayN_tfar(ray_o,N,i) = ray_i.ray.tfar;
     RTCRayN_time(ray_o,N,i) = ray_i.ray.time;
     RTCRayN_mask(ray_o,N,i) = ray_i.ray.mask;
+    RTCRayN_id(ray_o,N,i) = ray_i.ray.id;
     RTCHitN* hit_o = RTCRayHitN_HitN(rayhit_o,N);
-    RTCHitN_instID(hit_o,N,i,0) = ray_i.hit.instID[0];
     RTCHitN_geomID(hit_o,N,i) = ray_i.hit.geomID;
     RTCHitN_primID(hit_o,N,i) = ray_i.hit.primID;
     RTCHitN_u(hit_o,N,i) = ray_i.hit.u;
@@ -347,6 +378,9 @@ namespace embree
     RTCHitN_Ng_x(hit_o,N,i) = ray_i.hit.Ng_x;
     RTCHitN_Ng_y(hit_o,N,i) = ray_i.hit.Ng_y;
     RTCHitN_Ng_z(hit_o,N,i) = ray_i.hit.Ng_z;
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      RTCHitN_instID(hit_o,N,i,l) = ray_i.hit.instID[l];
   }
 
   __forceinline RTCRayHit getRay(RTCRayHit4& ray_i, size_t i)
@@ -362,7 +396,7 @@ namespace embree
     ray_o.ray.tfar = ray_i.ray.tfar[i];
     ray_o.ray.time = ray_i.ray.time[i];
     ray_o.ray.mask = ray_i.ray.mask[i];
-    ray_o.hit.instID[0] = ray_i.hit.instID[0][i];
+    ray_o.ray.id = ray_i.ray.id[i];
     ray_o.hit.geomID = ray_i.hit.geomID[i];
     ray_o.hit.primID = ray_i.hit.primID[i];
     ray_o.hit.u = ray_i.hit.u[i];
@@ -370,6 +404,10 @@ namespace embree
     ray_o.hit.Ng_x = ray_i.hit.Ng_x[i];
     ray_o.hit.Ng_y = ray_i.hit.Ng_y[i];
     ray_o.hit.Ng_z = ray_i.hit.Ng_z[i];
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l] = ray_i.hit.instID[l][i];
+
     return ray_o;
   }
 
@@ -386,7 +424,7 @@ namespace embree
     ray_o.ray.tfar = ray_i.ray.tfar[i];
     ray_o.ray.time = ray_i.ray.time[i];
     ray_o.ray.mask = ray_i.ray.mask[i];
-    ray_o.hit.instID[0] = ray_i.hit.instID[0][i];
+    ray_o.ray.id = ray_i.ray.id[i];
     ray_o.hit.geomID = ray_i.hit.geomID[i];
     ray_o.hit.primID = ray_i.hit.primID[i];
     ray_o.hit.u = ray_i.hit.u[i];
@@ -394,6 +432,10 @@ namespace embree
     ray_o.hit.Ng_x = ray_i.hit.Ng_x[i];
     ray_o.hit.Ng_y = ray_i.hit.Ng_y[i];
     ray_o.hit.Ng_z = ray_i.hit.Ng_z[i];
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l] = ray_i.hit.instID[l][i];
+
     return ray_o;
   }
 
@@ -410,7 +452,7 @@ namespace embree
     ray_o.ray.tfar = ray_i.ray.tfar[i];
     ray_o.ray.time = ray_i.ray.time[i];
     ray_o.ray.mask = ray_i.ray.mask[i];
-    ray_o.hit.instID[0] = ray_i.hit.instID[0][i];
+    ray_o.ray.id = ray_i.ray.id[i];
     ray_o.hit.geomID = ray_i.hit.geomID[i];
     ray_o.hit.primID = ray_i.hit.primID[i];
     ray_o.hit.u = ray_i.hit.u[i];
@@ -418,6 +460,9 @@ namespace embree
     ray_o.hit.Ng_x = ray_i.hit.Ng_x[i];
     ray_o.hit.Ng_y = ray_i.hit.Ng_y[i];
     ray_o.hit.Ng_z = ray_i.hit.Ng_z[i];
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l] = ray_i.hit.instID[l][i];
     return ray_o;
   }
 
@@ -436,7 +481,7 @@ namespace embree
     ray_o.ray.tfar  = RTCRayN_tfar(ray_i,N,i);
     ray_o.ray.time = RTCRayN_time(ray_i,N,i);
     ray_o.ray.mask = RTCRayN_mask(ray_i,N,i);
-    ray_o.hit.instID[0] = RTCHitN_instID(hit_i,N,i,0);
+    ray_o.ray.id = RTCRayN_id(ray_i,N,i);
     ray_o.hit.geomID = RTCHitN_geomID(hit_i,N,i);
     ray_o.hit.primID = RTCHitN_primID(hit_i,N,i);
     ray_o.hit.u = RTCHitN_u(hit_i,N,i);
@@ -444,6 +489,10 @@ namespace embree
     ray_o.hit.Ng_x = RTCHitN_Ng_x(hit_i,N,i);
     ray_o.hit.Ng_y = RTCHitN_Ng_y(hit_i,N,i);
     ray_o.hit.Ng_z = RTCHitN_Ng_z(hit_i,N,i);
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      ray_o.hit.instID[l] = RTCHitN_instID(hit_i, N, i, l);
+
     return ray_o;
   }
 
@@ -692,7 +741,6 @@ namespace embree
     rayp.ray.mask = &RTCRayN_mask(ray, N, 0);
     rayp.ray.id = &RTCRayN_id(ray, N, 0);
     rayp.ray.flags = &RTCRayN_flags(ray, N, 0);
-    rayp.hit.instID[0] = &RTCHitN_instID(hit, N, 0, 0);
     rayp.hit.geomID = &RTCHitN_geomID(hit, N, 0);
     rayp.hit.primID = &RTCHitN_primID(hit, N, 0);
     rayp.hit.u = &RTCHitN_u(hit, N, 0);
@@ -700,6 +748,9 @@ namespace embree
     rayp.hit.Ng_x = &RTCHitN_Ng_x(hit, N, 0);
     rayp.hit.Ng_y = &RTCHitN_Ng_y(hit, N, 0);
     rayp.hit.Ng_z = &RTCHitN_Ng_z(hit, N, 0);
+
+    for (unsigned l = 0; l < RTC_MAX_INSTANCE_LEVEL_COUNT; ++l)
+      rayp.hit.instID[l] = &RTCHitN_instID(hit, N, 0, l);
     
     switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
     case VARIANT_INTERSECT: rtcIntersectNp(scene, context, &rayp, N); break;
@@ -710,11 +761,16 @@ namespace embree
     for (unsigned int j = 0; j < N; j++) rays[j] = getRay(rayhit, N, j);
   }
 	
-  __noinline void IntersectWithModeInternal(IntersectMode mode, IntersectVariant ivariant, RTCScene scene, RTCRayHit* rays, unsigned int N)
+  __noinline void IntersectWithModeInternal(IntersectMode mode, IntersectVariant ivariant, RTCScene scene, RTCRayHit* rays, unsigned int N,
+      RTCIntersectContext* context)
   {
-    RTCIntersectContext context;
-    rtcInitIntersectContext(&context);
-    context.flags = ((ivariant & VARIANT_COHERENT_INCOHERENT_MASK) == VARIANT_COHERENT) ? RTC_INTERSECT_CONTEXT_FLAG_COHERENT :  RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT;
+    RTCIntersectContext _context;
+    if (!context)
+    {
+      rtcInitIntersectContext(&_context);
+      context = &_context;
+    }
+    context->flags = ((ivariant & VARIANT_COHERENT_INCOHERENT_MASK) == VARIANT_COHERENT) ? RTC_INTERSECT_CONTEXT_FLAG_COHERENT :  RTC_INTERSECT_CONTEXT_FLAG_INCOHERENT;
 
     switch (mode) 
     {
@@ -723,8 +779,8 @@ namespace embree
     case MODE_INTERSECT1:
     {
       switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-      case VARIANT_INTERSECT: for (size_t i=0; i<N; i++) rtcIntersect1(scene,&context,&rays[i]); break;
-      case VARIANT_OCCLUDED : for (size_t i=0; i<N; i++) rtcOccluded1 (scene,&context,(RTCRay*)&rays[i]); break;
+      case VARIANT_INTERSECT: for (size_t i=0; i<N; i++) rtcIntersect1(scene,context,&rays[i]); break;
+      case VARIANT_OCCLUDED : for (size_t i=0; i<N; i++) rtcOccluded1 (scene,context,(RTCRay*)&rays[i]); break;
       default: assert(false);
       }
       break;
@@ -741,8 +797,8 @@ namespace embree
         for (size_t j=0; j<M; j++) setRay(ray4,j,rays[i+j]);
         for (size_t j=M; j<4; j++) setRay(ray4,j,makeRay(zero,zero,pos_inf,neg_inf));
         switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-        case VARIANT_INTERSECT: rtcIntersect4(valid,scene,&context,&ray4); break;
-        case VARIANT_OCCLUDED : rtcOccluded4 (valid,scene,&context,(RTCRay4*)&ray4); break;
+        case VARIANT_INTERSECT: rtcIntersect4(valid,scene,context,&ray4); break;
+        case VARIANT_OCCLUDED : rtcOccluded4 (valid,scene,context,(RTCRay4*)&ray4); break;
         default: assert(false);
         }
         for (size_t j=0; j<M; j++) rays[i+j] = getRay(ray4,j);
@@ -760,8 +816,8 @@ namespace embree
         for (size_t j=0; j<M; j++) setRay(ray8,j,rays[i+j]);
         for (size_t j=M; j<8; j++) setRay(ray8,j,makeRay(zero,zero,pos_inf,neg_inf));
         switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-        case VARIANT_INTERSECT: rtcIntersect8(valid,scene,&context,&ray8); break;
-        case VARIANT_OCCLUDED : rtcOccluded8 (valid,scene,&context,(RTCRay8*)&ray8); break;
+        case VARIANT_INTERSECT: rtcIntersect8(valid,scene,context,&ray8); break;
+        case VARIANT_OCCLUDED : rtcOccluded8 (valid,scene,context,(RTCRay8*)&ray8); break;
         default: assert(false);
         }
         for (size_t j=0; j<M; j++) rays[i+j] = getRay(ray8,j);
@@ -779,8 +835,8 @@ namespace embree
         for (size_t j=0; j<M ; j++) setRay(ray16,j,rays[i+j]);
         for (size_t j=M; j<16; j++) setRay(ray16,j,makeRay(zero,zero,pos_inf,neg_inf));
         switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-        case VARIANT_INTERSECT: rtcIntersect16(valid,scene,&context,&ray16); break;
-        case VARIANT_OCCLUDED : rtcOccluded16 (valid,scene,&context,(RTCRay16*)&ray16); break;
+        case VARIANT_INTERSECT: rtcIntersect16(valid,scene,context,&ray16); break;
+        case VARIANT_OCCLUDED : rtcOccluded16 (valid,scene,context,(RTCRay16*)&ray16); break;
         default: assert(false);
         }
         for (size_t j=0; j<M; j++) rays[i+j] = getRay(ray16,j);
@@ -790,8 +846,8 @@ namespace embree
     case MODE_INTERSECT1M: 
     {
       switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-      case VARIANT_INTERSECT: rtcIntersect1M(scene,&context,rays,N,sizeof(RTCRayHit)); break;
-      case VARIANT_OCCLUDED : rtcOccluded1M (scene,&context,(RTCRay*)rays,N,sizeof(RTCRayHit)); break;
+      case VARIANT_INTERSECT: rtcIntersect1M(scene,context,rays,N,sizeof(RTCRayHit)); break;
+      case VARIANT_OCCLUDED : rtcOccluded1M (scene,context,(RTCRay*)rays,N,sizeof(RTCRayHit)); break;
       default: assert(false);
       }
       break;
@@ -802,40 +858,41 @@ namespace embree
       RTCRayHit* rptrs[1024];
       for (size_t i=0; i<N; i++) rptrs[i] = &rays[i];
       switch (ivariant & VARIANT_INTERSECT_OCCLUDED_MASK) {
-      case VARIANT_INTERSECT: rtcIntersect1Mp(scene,&context,rptrs,N); break;
-      case VARIANT_OCCLUDED : rtcOccluded1Mp (scene,&context,(RTCRay**)rptrs,N); break;
+      case VARIANT_INTERSECT: rtcIntersect1Mp(scene,context,rptrs,N); break;
+      case VARIANT_OCCLUDED : rtcOccluded1Mp (scene,context,(RTCRay**)rptrs,N); break;
       default: assert(false);
       }
       break;
     }
     case MODE_INTERSECTNM1: {
-      IntersectWithNMMode<1>(ivariant,scene,&context,rays,N);
+      IntersectWithNMMode<1>(ivariant,scene,context,rays,N);
       break;
     }
     case MODE_INTERSECTNM3: {
-      IntersectWithNMMode<3>(ivariant,scene,&context,rays,N);
+      IntersectWithNMMode<3>(ivariant,scene,context,rays,N);
       break;
     }
     case MODE_INTERSECTNM4: {
-      IntersectWithNMMode<4>(ivariant,scene,&context,rays,N);
+      IntersectWithNMMode<4>(ivariant,scene,context,rays,N);
       break;
     }
     case MODE_INTERSECTNM8: {
-      IntersectWithNMMode<8>(ivariant,scene,&context,rays,N);
+      IntersectWithNMMode<8>(ivariant,scene,context,rays,N);
       break;
     }
     case MODE_INTERSECTNM16: {
-      IntersectWithNMMode<16>(ivariant,scene,&context,rays,N);
+      IntersectWithNMMode<16>(ivariant,scene,context,rays,N);
       break;
     }
     case MODE_INTERSECTNp: {
-	  IntersectWithNpMode(ivariant, scene, &context, rays, N);
+	  IntersectWithNpMode(ivariant, scene, context, rays, N);
       break;
     }
     }
   }
 
-  void IntersectWithMode(IntersectMode mode, IntersectVariant ivariant, RTCScene scene, RTCRayHit* rays, unsigned int N)
+  void IntersectWithMode(IntersectMode mode, IntersectVariant ivariant, RTCScene scene, RTCRayHit* rays, unsigned int N,
+      RTCIntersectContext* context = nullptr)
   {
     /* verify occluded result against intersect */
     if ((ivariant & VARIANT_INTERSECT_OCCLUDED) == VARIANT_INTERSECT_OCCLUDED)
@@ -846,8 +903,8 @@ namespace embree
         valid[i] = rays[i].ray.tnear <= rays[i].ray.tfar;
         rays2[i] = rays[i];
       }
-      IntersectWithModeInternal(mode,IntersectVariant(ivariant & ~VARIANT_OCCLUDED),scene,rays,N);
-      IntersectWithModeInternal(mode,IntersectVariant(ivariant & ~VARIANT_INTERSECT),scene,rays2.data(),N);
+      IntersectWithModeInternal(mode,IntersectVariant(ivariant & ~VARIANT_OCCLUDED),scene,rays,N, context);
+      IntersectWithModeInternal(mode,IntersectVariant(ivariant & ~VARIANT_INTERSECT),scene,rays2.data(),N, context);
       for (size_t i=0; i<N; i++)
       {
         if (valid[i] && ((rays[i].hit.geomID == RTC_INVALID_GEOMETRY_ID) != (rays2[i].ray.tfar != float(neg_inf)))) {
@@ -856,7 +913,7 @@ namespace embree
       }
     }
     else
-      IntersectWithModeInternal(mode,ivariant,scene,rays,N);
+      IntersectWithModeInternal(mode,ivariant,scene,rays,N, context);
   }
 
   enum GeometryType
@@ -869,10 +926,12 @@ namespace embree
     GRID_MESH_MB,
     SUBDIV_MESH,
     SUBDIV_MESH_MB,
-    HAIR_GEOMETRY,
-    HAIR_GEOMETRY_MB,
-    CURVE_GEOMETRY,
-    CURVE_GEOMETRY_MB,
+    BEZIER_GEOMETRY,
+    BEZIER_GEOMETRY_MB,
+    BSPLINE_GEOMETRY,
+    BSPLINE_GEOMETRY_MB,
+    CATMULL_GEOMETRY,
+    CATMULL_GEOMETRY_MB,
     LINE_GEOMETRY,
     LINE_GEOMETRY_MB,
     SPHERE_GEOMETRY,
@@ -894,10 +953,12 @@ namespace embree
     case GRID_MESH_MB     : return "grids_mb";
     case SUBDIV_MESH      : return "subdivs";
     case SUBDIV_MESH_MB   : return "subdivs_mb";
-    case HAIR_GEOMETRY    : return "hair";
-    case HAIR_GEOMETRY_MB : return "hair_mb";
-    case CURVE_GEOMETRY   : return "curves";
-    case CURVE_GEOMETRY_MB: return "curves_mb";
+    case BEZIER_GEOMETRY  : return "bezier";
+    case BEZIER_GEOMETRY_MB: return "bezier_mb";
+    case BSPLINE_GEOMETRY : return "bspline";
+    case BSPLINE_GEOMETRY_MB: return "bspline_mb";
+    case CATMULL_GEOMETRY : return "catmull";
+    case CATMULL_GEOMETRY_MB : return "catmull_mb";
     case LINE_GEOMETRY    : return "lines";
     case LINE_GEOMETRY_MB : return "lines_mb";
     case SPHERE_GEOMETRY          : return "spheres";
